@@ -79,8 +79,8 @@ class btchip:
 		self.dongle = dongle
 		self.needKeyCache = False
 		try:
-			firmware = self.getFirmwareVersion()['version'].split(".")
-			self.multiOutputSupported = int(firmware[0]) >= 1 and int(firmware[1]) >= 1 and int(firmware[2]) >= 4
+			firmware = self.getFirmwareVersion()['version']
+			self.multiOutputSupported = tuple(map(int, (firmware.split(".")))) >= (1, 1, 4)
 			if self.multiOutputSupported:
 				self.scriptBlockLength = 50
 			else:
@@ -114,12 +114,12 @@ class btchip:
 				return e.sw - 0x63c0
 			raise e
 
-	def getWalletPublicKey(self, path, showOnScreen=False, segwit=False, segwitNative=False):
+	def getWalletPublicKey(self, path, showOnScreen=False, segwit=False, segwitNative=False, cashAddr=False):
 		result = {}
 		donglePath = parse_bip32_path(path)
 		if self.needKeyCache:
 			self.resolvePublicKeysInPath(path)			
-		apdu = [ self.BTCHIP_CLA, self.BTCHIP_INS_GET_WALLET_PUBLIC_KEY, 0x01 if showOnScreen else 0x00, 0x02 if segwitNative else 0x01 if segwit else 0x00, len(donglePath) ]
+		apdu = [ self.BTCHIP_CLA, self.BTCHIP_INS_GET_WALLET_PUBLIC_KEY, 0x01 if showOnScreen else 0x00, 0x03 if cashAddr else 0x02 if segwitNative else 0x01 if segwit else 0x00, len(donglePath) ]
 		apdu.extend(donglePath)
 		response = self.dongle.exchange(bytearray(apdu))
 		offset = 0
@@ -149,7 +149,7 @@ class btchip:
 			apdu.extend(params)
 			self.dongle.exchange(bytearray(apdu))
 			offset = 0
-			while (offset < len(trinput.script)):
+			while True:
 				blockLength = 251
 				if ((offset + blockLength) < len(trinput.script)):
 					dataLength = blockLength
@@ -162,6 +162,8 @@ class btchip:
 				apdu.extend(params)
 				self.dongle.exchange(bytearray(apdu))
 				offset += dataLength
+				if (offset >= len(trinput.script)):
+					break
 		# Number of outputs
 		apdu = [ self.BTCHIP_CLA, self.BTCHIP_INS_GET_TRUSTED_INPUT, 0x80, 0x00 ]
 		params = []
@@ -197,7 +199,7 @@ class btchip:
 		result['value'] = response
 		return result
 
-	def startUntrustedTransaction(self, newTransaction, inputIndex, outputList, redeemScript, version=0x01):
+	def startUntrustedTransaction(self, newTransaction, inputIndex, outputList, redeemScript, version=0x01, cashAddr=False, continueSegwit=False):
 		# Start building a fake transaction with the passed inputs
 		segwit = False
 		if newTransaction:
@@ -207,11 +209,11 @@ class btchip:
 					break
 		if newTransaction:
 			if segwit:
-				p2 = 0x02
+				p2 = 0x03 if cashAddr else 0x02
 			else:
 				p2 = 0x00
 		else:
-				p2 = 0x80
+				p2 = 0x10 if continueSegwit else 0x80
 		apdu = [ self.BTCHIP_CLA, self.BTCHIP_INS_HASH_INPUT_START, 0x00, p2 ]
 		params = bytearray([version, 0x00, 0x00, 0x00])
 		writeVarint(len(outputList), params)
@@ -228,10 +230,10 @@ class btchip:
 			apdu = [ self.BTCHIP_CLA, self.BTCHIP_INS_HASH_INPUT_START, 0x80, 0x00 ]
 			params = []
 			script = bytearray(redeemScript)
-			if ('witness' in passedOutput) and passedOutput['witness']:
-				params.append(0x02)
-			elif ('trustedInput' in passedOutput) and passedOutput['trustedInput']:
+			if ('trustedInput' in passedOutput) and passedOutput['trustedInput']:
 				params.append(0x01)
+			elif ('witness' in passedOutput) and passedOutput['witness']:
+				params.append(0x02)
 			else:
 				params.append(0x00)
 			if ('trustedInput' in passedOutput) and passedOutput['trustedInput']:
